#include "net_thread.h"
#include "FreeRTOS_IP.h"
#include "FreeRTOS_IP_Private.h"
#include "FreeRTOS_Sockets.h"
#include "common_utils.h"
#include "usr_app.h"
#include "server.h"
#include "SimStreamer.h"


/* DHCP populates these IP address, Sub net mask and Gateway Address. So start with this is zeroed out values
* The MAC address is Test MAC address.
*/
static  uint8_t ucMACAddress[ 6 ]       = {0x00, 0x11, 0x22, 0x33, 0x44, 0x55};
static  uint8_t ucIPAddress[ 4 ]        = {RESET_VALUE};
static  uint8_t ucNetMask[ 4 ]          = {RESET_VALUE};
static  uint8_t ucGatewayAddress[ 4 ]   = {RESET_VALUE};
static  uint8_t ucDNSServerAddress[ 4 ] = {RESET_VALUE};

extern NetworkAddressingParameters_t xNetworkAddressing;
NetworkAddressingParameters_t xNd = {RESET_VALUE, RESET_VALUE, RESET_VALUE, RESET_VALUE, RESET_VALUE};

uint32_t  dhcp_in_use   = RESET_VALUE;
ping_data_t ping_data   = {RESET_VALUE, RESET_VALUE, RESET_VALUE};

static rm_freertos_plus_fat_device_t FLASH;        /* Device structure contains device details */
FF_Disk_t QSPI_disk;

/* Private Function Declaration */
void reset_device(void);

/*******************************************************************************************************************//**
 * @brief      Generates 32 bit Random number
 * @param[in]  void
 * @retval     Random Number
 **********************************************************************************************************************/
uint32_t ulRand()
{
    /* example of a 32-bit random number generator.
     * Here rand() returns a 15-bit number. so create 32 bit Random number using 15 bit rand()
     */
    uint32_t ulResult =
        ((((uint32_t) rand()) & 0x7fffuL)       ) |
        ((((uint32_t) rand()) & 0x7fffuL) << 15 ) |
        ((((uint32_t) rand()) & 0x0003uL) << 30 );
    return ulResult;
}

 /*******************************************************************************************************************//**
 * @brief      Generates 32 sequence number
 * @param[in]  Source IP, Source Port, Destination IP and Destination Port.
 * @retval     Sequence Number
 **********************************************************************************************************************/

uint32_t ulApplicationGetNextSequenceNumber(uint32_t ulSourceAddress, uint16_t usSourcePort, uint32_t ulDestinationAddress,
                                            uint16_t usDestinationPort)
{
    /* Here we need to get random number for the sequence number.
     * This is just for testing purpose, so software rand() is okay.
     * This can also be tied to the TRNG.
     */

     return((ulSourceAddress + ulDestinationAddress + usSourcePort + usDestinationPort) && ulRand());
}

/*******************************************************************************************************************//**
 * @brief     This function Initializes the FreeRTOS+FAT instance..
 * @param[IN]   None
 * @retval      Any Other Error code apart from FSP_SUCCESS on Unsuccessful operation.
 ***********************************************************************************************************************/
static fsp_err_t FLASH_init(void)
{
    fsp_err_t err = FSP_SUCCESS;
    int32_t ff_err = FF_ERR_NONE;

    memset(&FLASH, RESET_VALUE, sizeof(FLASH));

    /* Open FreeRTOS PLUS FAT  */
    err = RM_FREERTOS_PLUS_FAT_Open(&g_rm_freertos_plus_fat_ctrl, &g_rm_freertos_plus_fat_cfg);
    if (FSP_SUCCESS != err)
    {
        APP_ERR_PRINT ("\r\nFREERTOS PLUS FAT OPEN API failed\r\n");
        return err;
    }

    APP_PRINT ("\r\nMounting FAT Partition 0\r\n");
    /* Wait for SD Card connection  */

    /* Initialize the mass storage device.  This should not be done until the device is plugged in and initialized. */
    err = RM_FREERTOS_PLUS_FAT_MediaInit(&g_rm_freertos_plus_fat_ctrl, &FLASH);
    if (FSP_SUCCESS != err)
    {
        APP_ERR_PRINT ("\r\nFreeRTOS Plus FAT Media Init API failed\r\n");
        RM_FREERTOS_PLUS_FAT_Close(&g_rm_freertos_plus_fat_ctrl);
        return err;
    }

    /* Initialize one disk for each partition used in the application. */
    err = RM_FREERTOS_PLUS_FAT_DiskInit(&g_rm_freertos_plus_fat_ctrl, &g_rm_freertos_plus_fat_disk_cfg, &QSPI_disk);
    if (FSP_SUCCESS != err)
    {
        APP_ERR_PRINT ("\r\nFreeRTOS Plus FAT Disk Init API failed\r\n");
        RM_FREERTOS_PLUS_FAT_Close(&g_rm_freertos_plus_fat_ctrl);
        return err;
    }

    /* Mount each disk.  This assumes the disk is already partitioned and formatted. */
    ff_err = FF_Mount(&QSPI_disk, QSPI_disk.xStatus.bPartitionNumber);
    if (FF_ERR_NONE != ff_err)
    {
        //mount_failed = true;
        APP_ERR_PRINT ("\r\nFF_Mount API failed indicating Disk is not formatted\r\n");
        APP_ERR_PRINT ("First Format the disk\r\n");
    }
    else
    {
        /* Add the disk to the file system. */
        ff_err = FF_FS_Add("/", &QSPI_disk);
        if (!ff_err)
        {
            APP_ERR_PRINT ("\r\nFF_FS_Add API failed\r\n");
            /* Close the FREERTOS_PLUS_FAT_Close instance on safely ejecting */
            RM_FREERTOS_PLUS_FAT_Close(&g_rm_freertos_plus_fat_ctrl);
            /* Returning FSP error in case of failure */
            return ((fsp_err_t )FF_ERR_DEVICE_DRIVER_FAILED);
        }
    }

    /* Set this flag to let application know that SD Card is initialized */
    return err;
}

/*******************************************************************************************************************//**
 * @brief     This function displays the content in current working directory.
 * @param[IN]   None
 * @retval      Any Other Error code apart from FSP_SUCCESS on Unsuccessful operation.
 ***********************************************************************************************************************/
static void display_FILES(void)
{
    /* Buffer to store CWD content */
    char pcBuffer[50];
    uint16_t Frames;
    FF_FindData_t * pxFindStruct = NULL;

    memset(&pcBuffer, RESET_VALUE, sizeof(pcBuffer));

    /* Allocate the memory */
    pxFindStruct = (FF_FindData_t *) pvPortMalloc(sizeof(FF_FindData_t));

    if(pxFindStruct)
    {
        memset( pxFindStruct, 0x00, sizeof( FF_FindData_t ) );

        /* pcBuffer is returned, hence return value is not captured */
        ff_getcwd(pcBuffer, sizeof(pcBuffer));
        APP_PRINT("Listing content in QSPI: %s\r\n", pcBuffer);
        APP_PRINT("Frame | File Name | File Size\r\n\r\n");

        /* The first parameter to ff_findfist() is the directory being searched.  Do
            not add wild cards to the end of the directory name. */
        if(RESET_VALUE == ff_findfirst("/", pxFindStruct))
        {
            do
            {
                if(RESET_VALUE == (pxFindStruct->ucAttributes & 0x1C) && (!strcmp(strrchr(pxFindStruct->pcFileName, '.'), ".jpg")))
                {   //Everything not dir, system or volumeID but jpeg file is considered user file (frame) to be streamed
                    Frames = SimStreamer::loadJPEG(pxFindStruct);
                    if(Frames)
                        APP_PRINT("%3d    %s    %d bytes\r\n", Frames, pxFindStruct->pcFileName, pxFindStruct->ulFileSize);
                    else
                        APP_PRINT("---    %s    %d bytes\r\n", pxFindStruct->pcFileName, pxFindStruct->ulFileSize);
                }

                vTaskDelay(10);
            } while(RESET_VALUE == ff_findnext(pxFindStruct));
        }
        /* Release the allocated memory */
        vPortFree((void *) pxFindStruct);
    }
}

/*******************************************************************************************************************//**
* @brief      This is the User Thread for the EP.
* @param[in]  Thread specific parameters
* @retval     None
**********************************************************************************************************************/
void net_thread_entry(void *pvParameters)
{
    BaseType_t status = pdFALSE;
    fsp_err_t sts = FSP_SUCCESS;

    fsp_pack_version_t version = {RESET_VALUE};

    FSP_PARAMETER_NOT_USED (pvParameters);

    /* version get API for FLEX pack information */
    R_FSP_VersionGet (&version);

    /* Example Project information printed on the RTT */
    APP_PRINT (BANNER_INFO, EP_VERSION, version.major, version.minor, version.patch);

    APP_PRINT("\r\nFlash Setting up...\r\n");

    if((sts = FLASH_init()) != FSP_SUCCESS)
    {
        APP_ERR_PRINT("Flash Error : FreeRTOS_QSPI_Init Failed");
        APP_ERR_TRAP(sts);
    }

    display_FILES();

    /* Reset device */
    reset_device();

    APP_PRINT("\r\nNetwork Setting up...\r\n");

    /* FreeRTOS IP Initialization: This init initializes the IP stack  */
    status = FreeRTOS_IPInit(ucIPAddress, ucNetMask, ucGatewayAddress, ucDNSServerAddress, ucMACAddress);

    if(pdFALSE == status)
    {
        APP_ERR_PRINT("Network Error : FreeRTOS_IPInit Failed");
        APP_ERR_TRAP(status);
    }

    APP_PRINT("\r\nConnecting");
    /* Display the New IP credentials obtained from the DHCP server */

    while(!dhcp_in_use)
    {
        vTaskDelay(10);
        APP_PRINT(".");
    }
    updateDhcpResponseToUsr();
    print_ipconfig();
    APP_PRINT("\r\n\r\nNetwork Setup OK\r\n");

    APP_PRINT("running RTSP server\r\n");

    HandleClients();
}

/*******************************************************************************************************************//**
* @brief      This is the User Hook for the DHCP Response. xApplicationDHCPHook() is called by DHCP Client Code when DHCP
*             handshake messages are exchanged from the Server.
* @param[in]  Different Phases of DHCP Phases and the Offered IP Address
* @retval     Returns DHCP Answers.
**********************************************************************************************************************/
eDHCPCallbackAnswer_t xApplicationDHCPHook( eDHCPCallbackPhase_t eDHCPPhase,
                                            uint32_t ulIPAddress )
{
eDHCPCallbackAnswer_t eReturn = eDHCPContinue;
  /*
   * This hook is called in a couple of places during the DHCP process, as identified by the eDHCPPhase parameter.
   */
  switch( eDHCPPhase )
  {
    case eDHCPPhasePreDiscover  :
      /*
       *  A DHCP discovery is about to be sent out.  eDHCPContinue is returned to allow the discovery to go out.
       *  If eDHCPUseDefaults had been returned instead then the DHCP process would be stopped and the statically
       *  configured IP address would be used.
       *  If eDHCPStopNoChanges had been returned instead then the DHCP process would be stopped and whatever the
       *  current network configuration was would continue to be used.
       */
        break;

    case eDHCPPhasePreRequest  :
      /* An offer has been received from the DHCP server, and the offered IP address is passed in the ulIPAddress
       * parameter.
       */

        /*
         * The sub-domains don’t match, so continue with the DHCP process so the offered IP address is used.
         */
        /* Update the Structure, the DHCP state Machine is not updating this */
        xNetworkAddressing.ulDefaultIPAddress = ulIPAddress;
        dhcp_in_use = 1;
      break;

    default :
      /*
       * Cannot be reached, but set eReturn to prevent compiler warnings where compilers are disposed to generating one.
       */
        break;
  }

  return eReturn;
}

/*******************************************************************************************************************//**
 * @brief      Creates and prints the the IP configuration to display on the RTT console
 * @param[in]  void
 * @retval     None
 **********************************************************************************************************************/

void print_ipconfig(void)
{
    if(dhcp_in_use)
    {
        ucNetMask[3] = (uint8_t)((xNd.ulNetMask & 0xFF000000) >> 24);
        ucNetMask[2] = (uint8_t)((xNd.ulNetMask & 0x00FF0000) >> 16);
        ucNetMask[1] = (uint8_t)((xNd.ulNetMask & 0x0000FF00) >> 8);
        ucNetMask[0] = (uint8_t)(xNd.ulNetMask & 0x000000FF);

        ucGatewayAddress[3] = (uint8_t)((xNd.ulGatewayAddress & 0xFF000000)>> 24);;
        ucGatewayAddress[2] = (uint8_t)((xNd.ulGatewayAddress & 0x00FF0000)>> 16);
        ucGatewayAddress[1] = (uint8_t)((xNd.ulGatewayAddress & 0x0000FF00)>> 8);
        ucGatewayAddress[0] = (uint8_t)(xNd.ulGatewayAddress & 0x000000FF);

        ucDNSServerAddress[3] = (uint8_t)((xNd.ulDNSServerAddress & 0xFF000000)>> 24);
        ucDNSServerAddress[2] = (uint8_t)((xNd.ulDNSServerAddress & 0x00FF0000)>> 16);
        ucDNSServerAddress[1] = (uint8_t)((xNd.ulDNSServerAddress & 0x0000FF00)>> 8);
        ucDNSServerAddress[0] = (uint8_t)(xNd.ulDNSServerAddress & 0x000000FF);

        ucIPAddress[3] = (uint8_t)((xNd.ulDefaultIPAddress & 0xFF000000) >> 24);
        ucIPAddress[2] = (uint8_t)((xNd.ulDefaultIPAddress & 0x00FF0000) >> 16);
        ucIPAddress[1] = (uint8_t)((xNd.ulDefaultIPAddress & 0x0000FF00) >> 8);
        ucIPAddress[0] = (uint8_t)(xNd.ulDefaultIPAddress & 0x000000FF);
    }
    APP_PRINT("\r\nEthernet adapter for Renesas " KIT_NAME ":\r\n");

    APP_PRINT("\tDescription . . . . . . . . . . . : Renesas " KIT_NAME " Ethernet\r\n");
    APP_PRINT("\tPhysical Address. . . . . . . . . : %02x-%02x-%02x-%02x-%02x-%02x\r\n",
            ucMACAddress[0],ucMACAddress[1],ucMACAddress[2],ucMACAddress[3],ucMACAddress[4],ucMACAddress[5]);
    APP_PRINT("\tDHCP Enabled. . . . . . . . . . . : %s\r\n",dhcp_in_use?"Yes":"No");
    APP_PRINT("\tIPv4 Address. . . . . . . . . . . : %d.%d.%d.%d\r\n",ucIPAddress[0],ucIPAddress[1],ucIPAddress[2],ucIPAddress[3]);
    APP_PRINT("\tSubnet Mask . . . . . . . . . . . : %d.%d.%d.%d\r\n",ucNetMask[0],ucNetMask[1],ucNetMask[2],ucNetMask[3]);
    APP_PRINT("\tDefault Gateway . . . . . . . . . : %d.%d.%d.%d\r\n",ucGatewayAddress[0],ucGatewayAddress[1],ucGatewayAddress[2],ucGatewayAddress[3]);
    APP_PRINT("\tDNS Servers . . . . . . . . . . . : %d.%d.%d.%d\r\n",ucDNSServerAddress[0],ucDNSServerAddress[1],ucDNSServerAddress[2],ucDNSServerAddress[3]);
}

/*******************************************************************************************************************//**
 * @brief      This Function checks the Network status (Both Ethernet and IP Layer). If the Network is down
 *             the Application will not send any data on the network.
 * @param[in]  None
 * @retval     Network Status
 **********************************************************************************************************************/
uint32_t isNetworkUp(void)
{
    fsp_err_t  eth_link_status = FSP_ERR_NOT_OPEN;
    BaseType_t networkUp = pdFALSE;
    uint32_t network_status = (IP_LINK_UP | ETHERNET_LINK_UP);

    networkUp = FreeRTOS_IsNetworkUp();
    eth_link_status = R_ETHER_LinkProcess(g_ether.p_ctrl);

    if((FSP_SUCCESS == eth_link_status) && (pdTRUE == networkUp))
    {
        return network_status;
    }
    else
    {
        if(FSP_SUCCESS != eth_link_status)
        {
            network_status |= ETHERNET_LINK_DOWN;
        }
        else if(FSP_SUCCESS == eth_link_status)
        {
            network_status |= ETHERNET_LINK_UP;
        }

        if(pdTRUE != networkUp)
        {
             network_status |= IP_LINK_DOWN;
        }
        else if(pdTRUE == networkUp)
        {
             network_status |= IP_LINK_UP;
        }
        return network_status;
    }
}

/*******************************************************************************************************************//**
 * @brief      Update the DHCP info to the User data structure.
 * @param[in]  None
 * @retval     None
 **********************************************************************************************************************/
void updateDhcpResponseToUsr(void)
{
    if(dhcp_in_use)
    {
        memcpy(&xNd, &xNetworkAddressing, sizeof(xNd));
    }
}

/*******************************************************************************************************************//**
 * @brief      DHCP Hook function to populate the user defined Host name for the Kit.
 * @param[in]  None
 * @retval     Hostname
 **********************************************************************************************************************/
#if( ipconfigDHCP_REGISTER_HOSTNAME == 1 )
const char *pcApplicationHostnameHook(void)
{
    return KIT_NAME;
}
#endif

/*****************************************************************************************************************
 *  @brief      Resets the phy device
 *  @param[in]  None
 *  @retval     None
 ****************************************************************************************************************/
void reset_device()
{
    /* PHY reset Pin */
    //R_IOPORT_PinWrite(&g_ioport_ctrl, ET_nRST, BSP_IO_LEVEL_LOW);
    R_BSP_SoftwareDelay(DELAY_TIME, BSP_DELAY_UNITS_MICROSECONDS);
    R_IOPORT_PinWrite(&g_ioport_ctrl, ET_nRST, BSP_IO_LEVEL_HIGH);
    R_BSP_SoftwareDelay(DELAY_TIME, BSP_DELAY_UNITS_MICROSECONDS);
}


/*******************************************************************************************************************//**
* @brief      User Hook for the Ping Reply. vApplicationPingReplyHook() is called by the TCP/IP
*             stack when the stack receives a ping reply.
* @param[in]  Ping reply status and Identifier
* @retval     None
**********************************************************************************************************************/
void vApplicationPingReplyHook( ePingReplyStatus_t eStatus, uint16_t usIdentifier )
{
    (void)  usIdentifier;

    switch( eStatus )
    {
        /* A valid ping reply has been received */
        case eSuccess    :
            ping_data.received++;
            break;
            /* A reply was received but it was not valid. */
        case eInvalidData :
        default:
            ping_data.lost++;
            break;
    }
}
