#include "common_utils.h"

#define NUM_STRING_DESCRIPTOR               (7U)

/* Global variables */
extern uint8_t g_apl_device[];
extern uint8_t g_apl_configuration[];
extern uint8_t g_apl_hs_configuration[];
extern uint8_t g_apl_qualifier_descriptor[];
extern uint8_t *g_apl_string_table[];

const usb_descriptor_t g_usb_descriptor =
{
g_apl_device,                   /* Pointer to the device descriptor */
g_apl_configuration,            /* Pointer to the configuration descriptor for Full-speed */
g_apl_hs_configuration,         /* Pointer to the configuration descriptor for Hi-speed */
g_apl_qualifier_descriptor,     /* Pointer to the qualifier descriptor */
g_apl_string_table,             /* Pointer to the string descriptor table */
NUM_STRING_DESCRIPTOR
};

FSP_CPP_HEADER
void R_BSP_WarmStart(bsp_warm_start_event_t event);
FSP_CPP_FOOTER

static bool r_qspi_get_QE(void)
{
    /* Enter direct communication mode */
    R_QSPI->SFMCMD = 1U;

    R_QSPI->SFMCOM = 0x35;
    bool sts = (R_QSPI->SFMCOM >> 1) & 1;

    /* Close the SPI bus cycle. Reference section 39.10.3 "Generating the SPI Bus Cycle during Direct Communication"
     * in the RA6M3 manual R01UH0886EJ0100. */
    R_QSPI->SFMCMD = 1U;

    /* Exit direct communication mode */
    R_QSPI->SFMCMD = 0U;

    return sts;
}

static void r_qspi_set_QE(qspi_instance_ctrl_t * p_instance_ctrl)
{
    /* Enter direct communication mode */
    R_QSPI->SFMCMD = 1U;

    /* Enable write. */
    R_QSPI->SFMCOM = p_instance_ctrl->p_cfg->write_enable_command;

    /* Close the SPI bus cycle. Reference section 39.10.3 "Generating the SPI Bus Cycle during Direct Communication"
     * in the RA6M3 manual R01UH0886EJ0100. */
    R_QSPI->SFMCMD = 1U;

    /* Send erase command. */
    R_QSPI->SFMCOM = 0x31;

    /* Send address if this is not a chip erase command. */
    R_QSPI->SFMCOM = (uint8_t) (0x02);

    /* Close the SPI bus cycle. Reference section 39.10.3 "Generating the SPI Bus Cycle during Direct Communication"
     * in the RA6M3 manual R01UH0886EJ0100. */
    R_QSPI->SFMCMD = 1U;
}

fsp_err_t R_QSPI_QuadEnable(spi_flash_ctrl_t * p_ctrl)
{
    qspi_instance_ctrl_t * p_instance_ctrl = (qspi_instance_ctrl_t *) p_ctrl;

#if QSPI_CFG_PARAM_CHECKING_ENABLE
    FSP_ASSERT(NULL != p_instance_ctrl);
    FSP_ASSERT(NULL != p_status);
    FSP_ERROR_RETURN(QSPI_PRV_OPEN == p_instance_ctrl->open, FSP_ERR_NOT_OPEN);

    /* Do not enter direct communication mode from XIP mode.  Reference note in section 39.10.2 "Using Direct
     * Communication Mode" in the RA6M3 manual R01UH0886EJ0100. */
    FSP_ERROR_RETURN(0U == R_QSPI->SFMSDC_b.SFMXST, FSP_ERR_INVALID_MODE);
#endif

    /* Read device status. */
    if(!r_qspi_get_QE())
        r_qspi_set_QE(p_instance_ctrl);

    return FSP_SUCCESS;
}

/*******************************************************************************************************************//**
 * main() is generated by the RA Configuration editor and is used to generate threads if an RTOS is used.  This function
 * is called by main() when no RTOS is used.
 **********************************************************************************************************************/
void hal_entry(void)
{
    usb_status_t event;
    fsp_pack_version_t version = {RESET_VALUE};

    /* version get API for FLEX pack information */
    R_FSP_VersionGet(&version);

    /* Example Project information printed on the Console */
    APP_PRINT(BANNER_INFO, EP_VERSION, version.major, version.minor, version.patch);

    if( FSP_SUCCESS == R_QSPI_Open(&g_qspi_ctrl, &g_qspi_cfg))
    {
        R_QSPI_QuadEnable(&g_qspi_ctrl);
        R_QSPI_Close(&g_qspi_ctrl);
    }

    R_USB_Open(&g_basic_ctrl, &g_basic_cfg);

    while(true)
    {
        /* check for usb event */
        if( FSP_SUCCESS == R_USB_EventGet(&g_basic_ctrl, &event))
            switch(event)
            {
                case USB_STATUS_CONFIGURED :
                    APP_PRINT("USB MSD[%d] Configured Successfully\r\n", g_basic_cfg.module_number)
                break;

                case USB_STATUS_DETACH :
              //case USB_STATUS_SUSPEND :
                    APP_PRINT("USB MSD[%d] Removed Successfully\r\n", g_basic_cfg.module_number)
                break;

                default :
                break;
            }
    }

#if BSP_TZ_SECURE_BUILD
    /* Enter non-secure code */
    R_BSP_NonSecureEnter();
#endif
}

/*******************************************************************************************************************//**
 * This function is called at various points during the startup process.  This implementation uses the event that is
 * called right before main() to set up the pins.
 *
 * @param[in]  event    Where at in the start up process the code is currently at
 **********************************************************************************************************************/
void R_BSP_WarmStart(bsp_warm_start_event_t event)
{
    if (BSP_WARM_START_RESET == event)
    {
#if BSP_FEATURE_FLASH_LP_VERSION != 0

        /* Enable reading from data flash. */
        R_FACI_LP->DFLCTL = 1U;

        /* Would normally have to wait tDSTOP(6us) for data flash recovery. Placing the enable here, before clock and
         * C runtime initialization, should negate the need for a delay since the initialization will typically take more than 6us. */
#endif
    }

    if (BSP_WARM_START_POST_C == event)
    {
        /* C runtime environment and system clocks are setup. */

        /* Configure pins. */
        R_IOPORT_Open (&g_ioport_ctrl, g_ioport.p_cfg);
      #ifdef __USE_TTY__
        TTY_init();
      #endif
    }
}

#if BSP_TZ_SECURE_BUILD

BSP_CMSE_NONSECURE_ENTRY void template_nonsecure_callable ();

/* Trustzone Secure Projects require at least one nonsecure callable function in order to build (Remove this if it is not required to build). */
BSP_CMSE_NONSECURE_ENTRY void template_nonsecure_callable ()
{

}
#endif
